@@ -1,4 +1,12 @@
-from sqlalchemy import cast
+from typing import Any
+from typing import cast
+from typing import Dict
+from typing import List
+from typing import Optional
+from typing import Tuple
+from typing import TYPE_CHECKING
+from typing import Union
+
 from sqlalchemy import CheckConstraint
 from sqlalchemy import Column
 from sqlalchemy import ForeignKeyConstraint
@@ -6,7 +14,6 @@ from sqlalchemy import Index
 from sqlalchemy import MetaData
 from sqlalchemy import PrimaryKeyConstraint
 from sqlalchemy import schema as sql_schema
-from sqlalchemy import select
 from sqlalchemy import Table
 from sqlalchemy import types as sqltypes
 from sqlalchemy.events import SchemaEventTarget
@@ -15,12 +22,28 @@ from sqlalchemy.util import topological
 
 from ..util import exc
 from ..util.sqla_compat import _columns_for_constraint
+from ..util.sqla_compat import _copy
+from ..util.sqla_compat import _ensure_scope_for_ddl
 from ..util.sqla_compat import _fk_is_self_referential
+from ..util.sqla_compat import _insert_inline
 from ..util.sqla_compat import _is_type_bound
 from ..util.sqla_compat import _remove_column_from_collection
+from ..util.sqla_compat import _select
+
+if TYPE_CHECKING:
+    from typing import Literal
+
+    from sqlalchemy.engine import Dialect
+    from sqlalchemy.sql.elements import ColumnClause
+    from sqlalchemy.sql.elements import quoted_name
+    from sqlalchemy.sql.functions import Function
+    from sqlalchemy.sql.schema import Constraint
+    from sqlalchemy.sql.type_api import TypeEngine
+
+    from ..ddl.impl import DefaultImpl
 
 
-class BatchOperationsImpl(object):
+class BatchOperationsImpl:
     def __init__(
         self,
         operations,
@@ -59,14 +82,14 @@ class BatchOperationsImpl(object):
         self.batch = []
 
     @property
-    def dialect(self):
+    def dialect(self) -> "Dialect":
         return self.operations.impl.dialect
 
     @property
-    def impl(self):
+    def impl(self) -> "DefaultImpl":
         return self.operations.impl
 
-    def _should_recreate(self):
+    def _should_recreate(self) -> bool:
         if self.recreate == "auto":
             return self.operations.impl.requires_recreate_in_batch(self)
         elif self.recreate == "always":
@@ -74,51 +97,52 @@ class BatchOperationsImpl(object):
         else:
             return False
 
-    def flush(self):
+    def flush(self) -> None:
         should_recreate = self._should_recreate()
 
-        if not should_recreate:
-            for opname, arg, kw in self.batch:
-                fn = getattr(self.operations.impl, opname)
-                fn(*arg, **kw)
-        else:
-            if self.naming_convention:
-                m1 = MetaData(naming_convention=self.naming_convention)
+        with _ensure_scope_for_ddl(self.impl.connection):
+            if not should_recreate:
+                for opname, arg, kw in self.batch:
+                    fn = getattr(self.operations.impl, opname)
+                    fn(*arg, **kw)
             else:
-                m1 = MetaData()
+                if self.naming_convention:
+                    m1 = MetaData(naming_convention=self.naming_convention)
+                else:
+                    m1 = MetaData()
 
-            if self.copy_from is not None:
-                existing_table = self.copy_from
-                reflected = False
-            else:
-                existing_table = Table(
-                    self.table_name,
-                    m1,
-                    schema=self.schema,
-                    autoload=True,
-                    autoload_with=self.operations.get_bind(),
-                    *self.reflect_args,
-                    **self.reflect_kwargs
+                if self.copy_from is not None:
+                    existing_table = self.copy_from
+                    reflected = False
+                else:
+                    existing_table = Table(
+                        self.table_name,
+                        m1,
+                        schema=self.schema,
+                        autoload_with=self.operations.get_bind(),
+                        *self.reflect_args,
+                        **self.reflect_kwargs
+                    )
+                    reflected = True
+
+                batch_impl = ApplyBatchImpl(
+                    self.impl,
+                    existing_table,
+                    self.table_args,
+                    self.table_kwargs,
+                    reflected,
+                    partial_reordering=self.partial_reordering,
                 )
-                reflected = True
+                for opname, arg, kw in self.batch:
+                    fn = getattr(batch_impl, opname)
+                    fn(*arg, **kw)
 
-            batch_impl = ApplyBatchImpl(
-                existing_table,
-                self.table_args,
-                self.table_kwargs,
-                reflected,
-                partial_reordering=self.partial_reordering,
-            )
-            for opname, arg, kw in self.batch:
-                fn = getattr(batch_impl, opname)
-                fn(*arg, **kw)
+                batch_impl._create(self.impl)
 
-            batch_impl._create(self.impl)
-
-    def alter_column(self, *arg, **kw):
+    def alter_column(self, *arg, **kw) -> None:
         self.batch.append(("alter_column", arg, kw))
 
-    def add_column(self, *arg, **kw):
+    def add_column(self, *arg, **kw) -> None:
         if (
             "insert_before" in kw or "insert_after" in kw
         ) and not self._should_recreate():
@@ -128,43 +152,61 @@ class BatchOperationsImpl(object):
             )
         self.batch.append(("add_column", arg, kw))
 
-    def drop_column(self, *arg, **kw):
+    def drop_column(self, *arg, **kw) -> None:
         self.batch.append(("drop_column", arg, kw))
 
-    def add_constraint(self, const):
+    def add_constraint(self, const: "Constraint") -> None:
         self.batch.append(("add_constraint", (const,), {}))
 
-    def drop_constraint(self, const):
+    def drop_constraint(self, const: "Constraint") -> None:
         self.batch.append(("drop_constraint", (const,), {}))
 
     def rename_table(self, *arg, **kw):
         self.batch.append(("rename_table", arg, kw))
 
-    def create_index(self, idx):
+    def create_index(self, idx: "Index") -> None:
         self.batch.append(("create_index", (idx,), {}))
 
-    def drop_index(self, idx):
+    def drop_index(self, idx: "Index") -> None:
         self.batch.append(("drop_index", (idx,), {}))
 
+    def create_table_comment(self, table):
+        self.batch.append(("create_table_comment", (table,), {}))
+
+    def drop_table_comment(self, table):
+        self.batch.append(("drop_table_comment", (table,), {}))
+
     def create_table(self, table):
         raise NotImplementedError("Can't create table in batch mode")
 
     def drop_table(self, table):
         raise NotImplementedError("Can't drop table in batch mode")
 
+    def create_column_comment(self, column):
+        self.batch.append(("create_column_comment", (column,), {}))
 
-class ApplyBatchImpl(object):
+
+class ApplyBatchImpl:
     def __init__(
-        self, table, table_args, table_kwargs, reflected, partial_reordering=()
-    ):
+        self,
+        impl: "DefaultImpl",
+        table: "Table",
+        table_args: tuple,
+        table_kwargs: Dict[str, Any],
+        reflected: bool,
+        partial_reordering: tuple = (),
+    ) -> None:
+        self.impl = impl
         self.table = table  # this is a Table object
         self.table_args = table_args
         self.table_kwargs = table_kwargs
         self.temp_table_name = self._calc_temp_name(table.name)
-        self.new_table = None
+        self.new_table: Optional[Table] = None
 
         self.partial_reordering = partial_reordering  # tuple of tuples
-        self.add_col_ordering = ()  # tuple of tuples
+        self.add_col_ordering: Tuple[
+            Tuple[str, str], ...
+        ] = ()  # tuple of tuples
 
         self.column_transfers = OrderedDict(
             (c.name, {"expr": c}) for c in self.table.c
@@ -175,29 +217,36 @@ class ApplyBatchImpl(object):
         self._grab_table_elements()
 
     @classmethod
-    def _calc_temp_name(cls, tablename):
+    def _calc_temp_name(cls, tablename: "quoted_name") -> str:
         return ("_alembic_tmp_%s" % tablename)[0:50]
 
-    def _grab_table_elements(self):
+    def _grab_table_elements(self) -> None:
         schema = self.table.schema
-        self.columns = OrderedDict()
+        self.columns: Dict[str, "Column"] = OrderedDict()
         for c in self.table.c:
-            c_copy = c.copy(schema=schema)
+            c_copy = _copy(c, schema=schema)
             c_copy.unique = c_copy.index = False
             # ensure that the type object was copied,
             # as we may need to modify it in-place
             if isinstance(c.type, SchemaEventTarget):
                 assert c_copy.type is not c.type
             self.columns[c.name] = c_copy
-        self.named_constraints = {}
+        self.named_constraints: Dict[str, "Constraint"] = {}
         self.unnamed_constraints = []
-        self.indexes = {}
-        self.new_indexes = {}
+        self.col_named_constraints = {}
+        self.indexes: Dict[str, "Index"] = {}
+        self.new_indexes: Dict[str, "Index"] = {}
+
         for const in self.table.constraints:
             if _is_type_bound(const):
                 continue
-            elif self.reflected and isinstance(const, CheckConstraint):
-                # TODO: we are skipping reflected CheckConstraint because
+            elif (
+                self.reflected
+                and isinstance(const, CheckConstraint)
+                and not const.name
+            ):
+                # TODO: we are skipping unnamed reflected CheckConstraint
+                # because
                 # we have no way to determine _is_type_bound() for these.
                 pass
             elif const.name:
@@ -205,13 +254,19 @@ class ApplyBatchImpl(object):
             else:
                 self.unnamed_constraints.append(const)
 
+        if not self.reflected:
+            for col in self.table.c:
+                for const in col.constraints:
+                    if const.name:
+                        self.col_named_constraints[const.name] = (col, const)
+
         for idx in self.table.indexes:
             self.indexes[idx.name] = idx
 
         for k in self.table.kwargs:
             self.table_kwargs.setdefault(k, self.table.kwargs[k])
 
-    def _adjust_self_columns_for_partial_reordering(self):
+    def _adjust_self_columns_for_partial_reordering(self) -> None:
         pairs = set()
 
         col_by_idx = list(self.columns)
@@ -231,17 +286,17 @@ class ApplyBatchImpl(object):
         # this can happen if some columns were dropped and not removed
         # from existing_ordering.  this should be prevented already, but
         # conservatively making sure this didn't happen
-        pairs = [p for p in pairs if p[0] != p[1]]
+        pairs_list = [p for p in pairs if p[0] != p[1]]
 
         sorted_ = list(
-            topological.sort(pairs, col_by_idx, deterministic_order=True)
+            topological.sort(pairs_list, col_by_idx, deterministic_order=True)
         )
         self.columns = OrderedDict((k, self.columns[k]) for k in sorted_)
         self.column_transfers = OrderedDict(
             (k, self.column_transfers[k]) for k in sorted_
         )
 
-    def _transfer_elements_to_new_table(self):
+    def _transfer_elements_to_new_table(self) -> None:
         assert self.new_table is None, "Can only create new table once"
 
         m = MetaData()
@@ -269,6 +324,7 @@ class ApplyBatchImpl(object):
             if not const_columns.issubset(self.column_transfers):
                 continue
 
+            const_copy: "Constraint"
             if isinstance(const, ForeignKeyConstraint):
                 if _fk_is_self_referential(const):
                     # for self-referential constraint, refer to the
@@ -277,22 +333,25 @@ class ApplyBatchImpl(object):
                     # FK constraints from other tables; we assume SQLite
                     # no foreign keys just keeps the names unchanged, so
                     # when we rename back, they match again.
-                    const_copy = const.copy(
-                        schema=schema, target_table=self.table
+                    const_copy = _copy(
+                        const, schema=schema, target_table=self.table
                     )
                 else:
                     # "target_table" for ForeignKeyConstraint.copy() is
                     # only used if the FK is detected as being
                     # self-referential, which we are handling above.
-                    const_copy = const.copy(schema=schema)
+                    const_copy = _copy(const, schema=schema)
             else:
-                const_copy = const.copy(schema=schema, target_table=new_table)
+                const_copy = _copy(
+                    const, schema=schema, target_table=new_table
+                )
             if isinstance(const, ForeignKeyConstraint):
                 self._setup_referent(m, const)
             new_table.append_constraint(const_copy)
 
-    def _gather_indexes_from_both_tables(self):
-        idx = []
+    def _gather_indexes_from_both_tables(self) -> List["Index"]:
+        assert self.new_table is not None
+        idx: List[Index] = []
         idx.extend(self.indexes.values())
         for index in self.new_indexes.values():
             idx.append(
@@ -305,8 +364,12 @@ class ApplyBatchImpl(object):
             )
         return idx
 
-    def _setup_referent(self, metadata, constraint):
-        spec = constraint.elements[0]._get_colspec()
+    def _setup_referent(
+        self, metadata: "MetaData", constraint: "ForeignKeyConstraint"
+    ) -> None:
+        spec = constraint.elements[
+            0
+        ]._get_colspec()  # type:ignore[attr-defined]
         parts = spec.split(".")
         tname = parts[-2]
         if len(parts) == 3:
@@ -316,11 +379,15 @@ class ApplyBatchImpl(object):
 
         if tname != self.temp_table_name:
             key = sql_schema._get_table_key(tname, referent_schema)
+
+            def colspec(elem: Any):
+                return elem._get_colspec()
+
             if key in metadata.tables:
                 t = metadata.tables[key]
                 for elem in constraint.elements:
-                    colname = elem._get_colspec().split(".")[-1]
-                    if not t.c.contains_column(colname):
+                    colname = colspec(elem).split(".")[-1]
+                    if colname not in t.c:
                         t.append_column(Column(colname, sqltypes.NULLTYPE))
             else:
                 Table(
@@ -329,29 +396,30 @@ class ApplyBatchImpl(object):
                     *[
                         Column(n, sqltypes.NULLTYPE)
                         for n in [
-                            elem._get_colspec().split(".")[-1]
+                            colspec(elem).split(".")[-1]
                             for elem in constraint.elements
                         ]
                     ],
                     schema=referent_schema
                 )
 
-    def _create(self, op_impl):
+    def _create(self, op_impl: "DefaultImpl") -> None:
         self._transfer_elements_to_new_table()
 
-        op_impl.prep_table_for_batch(self.table)
+        op_impl.prep_table_for_batch(self, self.table)
+        assert self.new_table is not None
         op_impl.create_table(self.new_table)
 
         try:
             op_impl._exec(
-                self.new_table.insert(inline=True).from_select(
+                _insert_inline(self.new_table).from_select(
                     list(
                         k
                         for k, transfer in self.column_transfers.items()
                         if "expr" in transfer
                     ),
-                    select(
-                        [
+                    _select(
+                        *[
                             transfer["expr"]
                             for transfer in self.column_transfers.values()
                             if "expr" in transfer
@@ -376,23 +444,34 @@ class ApplyBatchImpl(object):
 
     def alter_column(
         self,
-        table_name,
-        column_name,
-        nullable=None,
-        server_default=False,
-        name=None,
-        type_=None,
-        autoincrement=None,
+        table_name: str,
+        column_name: str,
+        nullable: Optional[bool] = None,
+        server_default: Optional[Union["Function", str, bool]] = False,
+        name: Optional[str] = None,
+        type_: Optional["TypeEngine"] = None,
+        autoincrement: None = None,
+        comment: Union[str, "Literal[False]"] = False,
         **kw
-    ):
+    ) -> None:
         existing = self.columns[column_name]
-        existing_transfer = self.column_transfers[column_name]
+        existing_transfer: Dict[str, Any] = self.column_transfers[column_name]
         if name is not None and name != column_name:
             # note that we don't change '.key' - we keep referring
             # to the renamed column by its old key in _create().  neat!
             existing.name = name
             existing_transfer["name"] = name
 
+            # pop named constraints for Boolean/Enum for rename
+            if (
+                "existing_type" in kw
+                and isinstance(kw["existing_type"], SchemaEventTarget)
+                and kw["existing_type"].name  # type:ignore[attr-defined]
+            ):
+                self.named_constraints.pop(
+                    kw["existing_type"].name, None  # type:ignore[attr-defined]
+                )
+
         if type_ is not None:
             type_ = sqltypes.to_instance(type_)
             # old type is being discarded so turn off eventing
@@ -400,15 +479,15 @@ class ApplyBatchImpl(object):
             # erase the events set up by this type, but this is simpler.
             # we also ignore the drop_constraint that will come here from
             # Operations.implementation_for(alter_column)
+
             if isinstance(existing.type, SchemaEventTarget):
-                existing.type._create_events = (
-                    existing.type.create_constraint
+                existing.type._create_events = (  # type:ignore[attr-defined]
+                    existing.type.create_constraint  # type:ignore[attr-defined] # noqa
                 ) = False
 
-            if existing.type._type_affinity is not type_._type_affinity:
-                existing_transfer["expr"] = cast(
-                    existing_transfer["expr"], type_
-                )
+            self.impl.cast_for_batch_migrate(
+                existing, existing_transfer, type_
+            )
 
             existing.type = type_
 
@@ -423,13 +502,23 @@ class ApplyBatchImpl(object):
             if server_default is None:
                 existing.server_default = None
             else:
-                sql_schema.DefaultClause(server_default)._set_parent(existing)
+                sql_schema.DefaultClause(
+                    server_default
+                )._set_parent(  # type:ignore[attr-defined]
+                    existing
+                )
         if autoincrement is not None:
             existing.autoincrement = bool(autoincrement)
 
+        if comment is not False:
+            existing.comment = comment
+
     def _setup_dependencies_for_add_column(
-        self, colname, insert_before, insert_after
-    ):
+        self,
+        colname: str,
+        insert_before: Optional[str],
+        insert_after: Optional[str],
+    ) -> None:
         index_cols = self.existing_ordering
         col_indexes = {name: i for i, name in enumerate(index_cols)}
 
@@ -473,17 +562,24 @@ class ApplyBatchImpl(object):
             self.add_col_ordering += ((index_cols[-1], colname),)
 
     def add_column(
-        self, table_name, column, insert_before=None, insert_after=None, **kw
-    ):
+        self,
+        table_name: str,
+        column: "Column",
+        insert_before: Optional[str] = None,
+        insert_after: Optional[str] = None,
+        **kw
+    ) -> None:
         self._setup_dependencies_for_add_column(
             column.name, insert_before, insert_after
         )
         # we copy the column because operations.add_column()
         # gives us a Column that is part of a Table already.
-        self.columns[column.name] = column.copy(schema=self.table.schema)
+        self.columns[column.name] = _copy(column, schema=self.table.schema)
         self.column_transfers[column.name] = {}
 
-    def drop_column(self, table_name, column, **kw):
+    def drop_column(
+        self, table_name: str, column: Union["ColumnClause", "Column"], **kw
+    ) -> None:
         if column.name in self.table.primary_key.columns:
             _remove_column_from_collection(
                 self.table.primary_key.columns, column
@@ -492,7 +588,39 @@ class ApplyBatchImpl(object):
         del self.column_transfers[column.name]
         self.existing_ordering.remove(column.name)
 
-    def add_constraint(self, const):
+        # pop named constraints for Boolean/Enum for rename
+        if (
+            "existing_type" in kw
+            and isinstance(kw["existing_type"], SchemaEventTarget)
+            and kw["existing_type"].name  # type:ignore[attr-defined]
+        ):
+            self.named_constraints.pop(
+                kw["existing_type"].name, None  # type:ignore[attr-defined]
+            )
+
+    def create_column_comment(self, column):
+        """the batch table creation function will issue create_column_comment
+        on the real "impl" as part of the create table process.
+
+        That is, the Column object will have the comment on it already,
+        so when it is received by add_column() it will be a normal part of
+        the CREATE TABLE and doesn't need an extra step here.
+
+        """
+
+    def create_table_comment(self, table):
+        """the batch table creation function will issue create_table_comment
+        on the real "impl" as part of the create table process.
+
+        """
+
+    def drop_table_comment(self, table):
+        """the batch table creation function will issue drop_table_comment
+        on the real "impl" as part of the create table process.
+
+        """
+
+    def add_constraint(self, const: "Constraint") -> None:
         if not const.name:
             raise ValueError("Constraint must have a name")
         if isinstance(const, sql_schema.PrimaryKeyConstraint):
@@ -501,11 +629,18 @@ class ApplyBatchImpl(object):
 
         self.named_constraints[const.name] = const
 
-    def drop_constraint(self, const):
+    def drop_constraint(self, const: "Constraint") -> None:
         if not const.name:
             raise ValueError("Constraint must have a name")
         try:
-            const = self.named_constraints.pop(const.name)
+            if const.name in self.col_named_constraints:
+                col, const = self.col_named_constraints.pop(const.name)
+
+                for col_const in list(self.columns[col.name].constraints):
+                    if col_const.name == const.name:
+                        self.columns[col.name].constraints.remove(col_const)
+            else:
+                const = self.named_constraints.pop(cast(str, const.name))
         except KeyError:
             if _is_type_bound(const):
                 # type-bound constraints are only included in the new
@@ -519,10 +654,10 @@ class ApplyBatchImpl(object):
                 for col in const.columns:
                     self.columns[col.name].primary_key = False
 
-    def create_index(self, idx):
+    def create_index(self, idx: "Index") -> None:
         self.new_indexes[idx.name] = idx
 
-    def drop_index(self, idx):
+    def drop_index(self, idx: "Index") -> None:
         try:
             del self.indexes[idx.name]
         except KeyError:
