@@ -29,6 +29,7 @@ from ..util.compat import string_types
 from ..util.compat import text_type
 from ..util.sqla_compat import create_mock_engine
 from ..util.sqla_compat import sqla_14
+from ..util.sqla_compat import sqla_1x
 
 
 testing_config = configparser.ConfigParser()
@@ -36,7 +37,10 @@ testing_config.read(["test.cfg"])
 
 
 class TestBase(SQLAlchemyTestBase):
-    is_sqlalchemy_future = False
+    if sqla_1x:
+        is_sqlalchemy_future = False
+    else:
+        is_sqlalchemy_future = True
 
     @testing.fixture()
     def ops_context(self, migration_context):
@@ -245,6 +249,7 @@ class AlterColRoundTripFixture:
         ), "server defaults %r and %r didn't compare as equivalent" % (s1, s2)
 
     def tearDown(self):
+        sqla_compat._safe_rollback_connection_transaction(self.conn)
         with self.conn.begin():
             self.metadata.drop_all(self.conn)
         self.conn.close()
