@@ -1,6 +1,7 @@
 import io
 import os
 import sys
+from typing import Tuple
 
 from sqlalchemy.util import inspect_getfullargspec  # noqa
 from sqlalchemy.util.compat import inspect_formatargspec  # noqa
@@ -26,19 +27,18 @@ class EncodedIO(io.TextIOWrapper):
 
 if py39:
     from importlib import resources as importlib_resources
+    from importlib import metadata as importlib_metadata
+    from importlib.metadata import EntryPoint
 else:
     import importlib_resources  # type:ignore[no-redef] # noqa
-
-if py38:
-    from importlib import metadata as importlib_metadata
-else:
     import importlib_metadata  # type:ignore[no-redef] # noqa
+    from importlib_metadata import EntryPoint  # type:ignore # noqa
 
 
-def importlib_metadata_get(group):
+def importlib_metadata_get(group: str) -> Tuple[EntryPoint, ...]:
     ep = importlib_metadata.entry_points()
     if hasattr(ep, "select"):
-        return ep.select(group=group)
+        return ep.select(group=group)  # type:ignore[attr-defined]
     else:
         return ep.get(group, ())
 
